import numpy

from .._mesh._mesh import CellBlock, Mesh

__all__ = [
    "triangulate",
]


def triangulate(points, material="dfalt"):
    """
    Generate a triangulated mesh from cloud points.

    Parameters
    ----------
    points : array_like
        Cooordinates of points.
    material : str, optional, default 'dfalt'
        Default material name.

    Returns
    -------
    toughio.Mesh
        Output triangulated mesh.

    """
    try:
        from scipy.spatial import Delaunay
    except ImportError:
        raise ImportError("Triangulation requires scipy >= 0.9 to be installed.")

    if not isinstance(points, (list, tuple, numpy.ndarray)):
        raise TypeError()
    if numpy.ndim(points) != 2:
        raise ValueError()
    if numpy.shape(points)[1] not in {2, 3}:
        raise ValueError()
    if not isinstance(material, str):
        raise TypeError()

    points = numpy.asarray(points)
    connectivity = Delaunay(points).simplices
    cell_type = "triangle" if points.shape[1] == 2 else "tetra"
    cells = [CellBlock(cell_type, connectivity)]
    mesh = Mesh(points, cells)
    mesh.add_cell_data("material", numpy.ones(mesh.n_cells, dtype=int))
    mesh.field_data[material] = numpy.array([1, 3])

    return mesh



from __future__ import division, unicode_literals, with_statement

import struct
import time

import numpy

from ...__about__ import __version__ as version
from .._helpers import get_material_key
from .._mesh import Mesh

__all__ = [
    "read",
    "write",
]


meshio_only = {
    "zone": {
        "tetra": "tetra",
        "tetra10": "tetra",
        "pyramid": "pyramid",
        "pyramid13": "pyramid",
        "wedge": "wedge",
        "wedge12": "wedge",
        "wedge15": "wedge",
        "wedge18": "wedge",
        "hexahedron": "hexahedron",
        "hexahedron20": "hexahedron",
        "hexahedron24": "hexahedron",
        "hexahedron27": "hexahedron",
    },
    "face": {
        "triangle": "triangle",
        "triangle6": "triangle",
        "triangle7": "triangle",
        "quad": "quad",
        "quad8": "quad",
        "quad9": "quad",
    },
}


numnodes_to_meshio_type = {
    "zone": {4: "tetra", 5: "pyramid", 6: "wedge", 8: "hexahedron"},
    "face": {3: "triangle", 4: "quad"},
}


meshio_to_flac3d_type = {
    "triangle": "T3",
    "quad": "Q4",
    "tetra": "T4",
    "pyramid": "P5",
    "wedge": "W6",
    "hexahedron": "B8",
}


flac3d_to_meshio_order = {
    "triangle": [0, 1, 2],
    "quad": [0, 1, 2, 3],
    "tetra": [0, 1, 2, 3],
    "pyramid": [0, 1, 4, 2, 3],
    "wedge": [0, 1, 3, 2, 4, 5],
    "hexahedron": [0, 1, 4, 2, 3, 6, 7, 5],
}


meshio_to_flac3d_order = {
    "triangle": [0, 1, 2],
    "quad": [0, 1, 2, 3],
    "tetra": [0, 1, 2, 3],
    "pyramid": [0, 1, 3, 4, 2],
    "wedge": [0, 1, 3, 2, 4, 5],
    "hexahedron": [0, 1, 3, 4, 2, 7, 5, 6],
}


meshio_to_flac3d_order_2 = {
    "tetra": [0, 2, 1, 3],
    "pyramid": [0, 3, 1, 4, 2],
    "wedge": [0, 2, 3, 1, 5, 4],
    "hexahedron": [0, 3, 1, 4, 2, 5, 7, 6],
}


flag_to_numdim = {
    "zone": 3,
    "face": 2,
}


def read(filename):
    """Read FLAC3D f3grid grid file."""
    # Read a small block of the file to assess its type
    # See <http://code.activestate.com/recipes/173220/>
    with open(filename, "rb") as f:
        block = f.read(8)
        binary = b"\x00" in block

    mode = "rb" if binary else "r"
    with open(filename, mode) as f:
        out = read_buffer(f, binary)

    return out


def read_buffer(f, binary):
    """Read binary or ASCII file."""
    flags = {
        "Z": "zone",
        "F": "face",
        "ZGROUP": "zone",
        "FGROUP": "face",
    }

    points = []
    point_ids = {}
    cells = []
    field_data = {}

    # Zones and faces do not share the same cell ID pool in FLAC3D
    # i.e. a given cell ID can be assigned to a zone and a face concurrently
    mapper = {"zone": {}, "face": {}}
    slots = {"zone": set(), "face": set()}

    pidx = 0
    cidx = 0
    gidx = 0
    if binary:
        # Not sure what the first bytes represent, the format might be wrong
        # It does not seem to be useful anyway
        _ = struct.unpack("<2I", f.read(8))

        (num_nodes,) = struct.unpack("<I", f.read(4))
        for pidx in range(num_nodes):
            pid, point = _read_point(f, binary)
            points.append(point)
            point_ids[pid] = pidx

        for flag in ["zone", "face"]:
            (num_cells,) = struct.unpack("<I", f.read(4))
            for _ in range(num_cells):
                cid, cell = _read_cell(f, point_ids, binary)
                cells = _update_cells(cells, cell, flag)
                mapper[flag][cid] = [cidx]
                cidx += 1

            (num_groups,) = struct.unpack("<I", f.read(4))
            for _ in range(num_groups):
                name, slot, data = _read_group(f, binary)
                field_data, mapper[flag] = _update_field_data(
                    field_data, mapper[flag], data, name, gidx + 1, flag,
                )
                slots[flag] = _update_slots(slots[flag], slot)
                gidx += 1
    else:
        line = f.readline().rstrip().split()
        while line:
            if line[0] == "G":
                pid, point = _read_point(line, binary)
                points.append(point)
                point_ids[pid] = pidx
                pidx += 1
            elif line[0] in {"Z", "F"}:
                flag = flags[line[0]]
                cid, cell = _read_cell(line, point_ids, binary)
                cells = _update_cells(cells, cell, flag)
                mapper[flag][cid] = [cidx]
                cidx += 1
            elif line[0] in {"ZGROUP", "FGROUP"}:
                flag = flags[line[0]]
                name, slot, data = _read_group(f, binary, line)
                field_data, mapper[flag] = _update_field_data(
                    field_data, mapper[flag], data, name, gidx + 1, flag,
                )
                slots[flag] = _update_slots(slots[flag], slot)
                gidx += 1

            line = f.readline().rstrip().split()

    if field_data:
        num_cells = numpy.cumsum([len(c[1]) for c in cells])
        cell_data = numpy.zeros(num_cells[-1], dtype=int)
        for k, v in mapper.items():
            if not slots[k]:
                continue

            for cid, zid in v.values():
                cell_data[cid] = zid
        cell_data = {"flac3d:group": numpy.split(cell_data, num_cells[:-1])}
    else:
        cell_data = {}

    return Mesh(
        points=numpy.array(points),
        cells=[(k, numpy.array(v)[:, flac3d_to_meshio_order[k]]) for k, v in cells],
        cell_data=cell_data,
        field_data=field_data,
    )


def _read_point(buf_or_line, binary):
    """Read point coordinates."""
    if binary:
        pid, x, y, z = struct.unpack("<I3d", buf_or_line.read(28))
        point = [x, y, z]
    else:
        pid = int(buf_or_line[1])
        point = [float(l) for l in buf_or_line[2:]]

    return pid, point


def _read_cell(buf_or_line, point_ids, binary):
    """Read cell connectivity."""
    if binary:
        cid, num_verts = struct.unpack("<2I", buf_or_line.read(8))
        cell = struct.unpack("<{}I".format(num_verts), buf_or_line.read(4 * num_verts))
        is_b7 = num_verts == 7
    else:
        cid = int(buf_or_line[2])
        cell = buf_or_line[3:]
        is_b7 = buf_or_line[1] == "B7"

    cell = [point_ids[int(l)] for l in cell]
    if is_b7:
        cell.append(cell[-1])

    return cid, cell


def _read_group(buf_or_line, binary, line=None):
    """Read cell group."""
    if binary:
        # Group name
        (num_chars,) = struct.unpack("<H", buf_or_line.read(2))
        (name,) = struct.unpack("<{}s".format(num_chars), buf_or_line.read(num_chars))
        name = name.decode("utf-8")

        # Slot name
        (num_chars,) = struct.unpack("<H", buf_or_line.read(2))
        (slot,) = struct.unpack("<{}s".format(num_chars), buf_or_line.read(num_chars))
        slot = slot.decode("utf-8")

        # Zones
        (num_zones,) = struct.unpack("<I", buf_or_line.read(4))
        data = struct.unpack("<{}I".format(num_zones), buf_or_line.read(4 * num_zones))
    else:
        name = line[1].replace('"', "")
        data = []
        slot = "" if "SLOT" not in line else line[-1]

        i = buf_or_line.tell()
        line = buf_or_line.readline()
        while True:
            line = line.rstrip().split()
            if line and (line[0] not in {"*", "ZGROUP"}):
                data += [int(l) for l in line]
            else:
                buf_or_line.seek(i)
                break
            i = buf_or_line.tell()
            line = buf_or_line.readline()

    return name, slot, data


def _update_cells(cells, cell, flag):
    """Update cell list."""
    cell_type = numnodes_to_meshio_type[flag][len(cell)]
    if len(cells) > 0 and cell_type == cells[-1][0]:
        cells[-1][1].append(cell)
    else:
        cells.append((cell_type, [cell]))

    return cells


def _update_field_data(field_data, mapper, data, name, gidx, flag):
    """Update field data dict."""
    for cid in data:
        mapper[cid].append(gidx)
    field_data[name] = numpy.array([gidx, flag_to_numdim[flag]])

    return field_data, mapper


def _update_slots(slots, slot):
    """
    Update slot set.

    Only one slot is supported.

    """
    slots.add(slot)
    if len(slots) > 1:
        raise ValueError("Multiple slots are not supported")

    return slots


def write(filename, mesh, float_fmt=".16e", binary=False):
    """Write FLAC3D f3grid grid file."""
    if not any(c.type in meshio_only["zone"].keys() for c in mesh.cells):
        raise ValueError("FLAC3D format only supports 3D cells")

    # Pick out material
    material = None
    if mesh.cell_data:
        key = get_material_key(mesh.cell_data)
        if key:
            material = numpy.concatenate(mesh.cell_data[key])

    mode = "wb" if binary else "w"
    with open(filename, mode) as f:
        if binary:
            f.write(
                struct.pack("<2I", 1375135718, 3)
            )  # Don't know what these values represent
        else:
            f.write("* FLAC3D grid produced by toughio v{}\n".format(version))
            f.write("* {}\n".format(time.ctime()))

        _write_points(f, mesh.points, binary, float_fmt)
        for flag in ["zone", "face"]:
            _write_cells(f, mesh.points, mesh.cells, flag, binary)
            _write_groups(f, mesh.cells, material, mesh.field_data, flag, binary)


def _write_points(f, points, binary, float_fmt=None):
    """Write points coordinates."""
    if binary:
        f.write(struct.pack("<I", len(points)))
        for i, point in enumerate(points):
            f.write(struct.pack("<I3d", i + 1, *point))
    else:
        f.write("* GRIDPOINTS\n")
        for i, point in enumerate(points):
            fmt = "G\t{:8}\t" + "\t".join(3 * ["{:" + float_fmt + "}"]) + "\n"
            f.write(fmt.format(i + 1, *point))


def _write_cells(f, points, cells, flag, binary):
    """Write cells."""
    if flag == "zone":
        count = 0
        cells = _translate_zones(points, cells)
    else:
        count = sum(len(c[1]) for c in cells if c.type in meshio_only["zone"])
        cells = _translate_faces(cells)

    if binary:
        f.write(
            struct.pack(
                "<I", sum(len(c[1]) for c in cells if c[0] in meshio_only[flag])
            )
        )
        for _, cdata in cells:
            num_cells, num_verts = cdata.shape
            tmp = numpy.column_stack(
                (
                    numpy.arange(1, num_cells + 1) + count,
                    numpy.full(num_cells, num_verts),
                    cdata + 1,
                )
            )
            f.write(
                struct.pack("<{}I".format((num_verts + 2) * num_cells), *tmp.ravel())
            )
            count += num_cells
    else:
        flag_to_text = {
            "zone": ("ZONES", "Z"),
            "face": ("FACES", "F"),
        }

        f.write("* {}\n".format(flag_to_text[flag][0]))
        for ctype, cdata in cells:
            fmt = (
                "{} {{}} {{}} ".format(flag_to_text[flag][1])
                + " ".join(["{}"] * cdata.shape[1])
                + "\n"
            )
            for entry in cdata + 1:
                count += 1
                f.write(fmt.format(meshio_to_flac3d_type[ctype], count, *entry))


def _write_groups(f, cells, cell_data, field_data, flag, binary):
    """Write groups."""
    if cell_data is not None:
        groups, labels = _translate_groups(cells, cell_data, field_data, flag)

        if binary:
            slot = "Default".encode("utf-8")

            f.write(struct.pack("<I", len(groups)))
            for k in sorted(groups.keys()):
                num_chars, num_zones = len(labels[k]), len(groups[k])
                fmt = "<H{}sH7sI{}I".format(num_chars, num_zones)
                tmp = [
                    num_chars,
                    labels[k].encode("utf-8"),
                    7,
                    slot,
                    num_zones,
                ]
                tmp += groups[k].tolist()
                f.write(struct.pack(fmt, *tmp))
        else:
            flag_to_text = {
                "zone": "ZGROUP",
                "face": "FGROUP",
            }

            f.write("* {} GROUPS\n".format(flag.upper()))
            for k in sorted(groups.keys()):
                f.write('{} "{}"\n'.format(flag_to_text[flag], labels[k]))
                _write_table(f, groups[k])
    else:
        if binary:
            f.write(struct.pack("<I", 0))


def _translate_zones(points, cells):
    """
    Reorder toughio cells to FLAC3D zones.

    Four first points must form a right-handed coordinate system (outward normal
    vectors). Reorder corner points according to sign of scalar triple products.

    """
    # See <https://stackoverflow.com/a/42386330/353337>
    def slicing_summing(a, b, c):
        c0 = b[:, 1] * c[:, 2] - b[:, 2] * c[:, 1]
        c1 = b[:, 2] * c[:, 0] - b[:, 0] * c[:, 2]
        c2 = b[:, 0] * c[:, 1] - b[:, 1] * c[:, 0]
        return a[:, 0] * c0 + a[:, 1] * c1 + a[:, 2] * c2

    zones = []
    for key, idx in cells:
        if key not in meshio_only["zone"].keys():
            continue

        # Compute scalar triple products
        key = meshio_only["zone"][key]
        tmp = points[idx[:, meshio_to_flac3d_order[key][:4]].T]
        det = slicing_summing(tmp[1] - tmp[0], tmp[2] - tmp[0], tmp[3] - tmp[0])

        # Reorder corner points
        data = numpy.where(
            (det > 0)[:, None],
            idx[:, meshio_to_flac3d_order[key]],
            idx[:, meshio_to_flac3d_order_2[key]],
        )
        zones.append((key, data))

    return zones


def _translate_faces(cells):
    """Reorder toughio cells to FLAC3D faces."""
    faces = []
    for key, idx in cells:
        if key not in meshio_only["face"].keys():
            continue

        key = meshio_only["face"][key]
        data = idx[:, meshio_to_flac3d_order[key]]
        faces.append((key, data))

    return faces


def _translate_groups(cells, cell_data, field_data, flag):
    """Convert toughio cell_data to FLAC3D groups."""
    num_dims = numpy.concatenate(
        [numpy.full(len(c[1]), 2 if c[0] in meshio_only["face"] else 3) for c in cells]
    )
    groups = {
        k: numpy.nonzero(
            numpy.logical_and(cell_data == k, num_dims == flag_to_numdim[flag])
        )[0]
        + 1
        for k in numpy.unique(cell_data)
    }
    groups = {k: v for k, v in groups.items() if v.size}

    labels = {k: str(k) for k in groups.keys()}
    labels[0] = "None"
    if field_data:
        labels.update(
            {v[0]: k for k, v in field_data.items() if v[1] == flag_to_numdim[flag]}
        )

    return groups, labels


def _write_table(f, data, ncol=20):
    """Write group data table."""
    nrow = len(data) // ncol
    lines = numpy.split(data, numpy.full(nrow, ncol).cumsum())
    for line in lines:
        if len(line):
            f.write(" {}\n".format(" ".join([str(l) for l in line])))



import os

import helpers
import numpy
import pytest

import toughio

write_read = lambda output, writer_kws, reader_kws: helpers.write_read(
    "output",
    output,
    toughio.write_output,
    toughio.read_output,
    writer_kws=writer_kws,
    reader_kws=reader_kws,
)


@pytest.mark.parametrize(
    "filename, data_ref",
    [
        (
            "FOFT_A1912.csv",
            {
                "TIME": 4.393722000e9,
                "PRES": 1.8740899675005e8,
                "TEMP": 720.0,
                "SAT_G": 0.0,
                "SAT_L": 24.0,
            },
        ),
        (
            "FOFT_A1912_T2.csv",
            {
                "TIME": 3.06639400e9,
                "PRES": 1.83000721e8,
                "TEMP": 660.0,
                "SAT_G": 0.0,
                "SAT_L": 22.0,
            },
        ),
        (
            "GOFT_A1162.csv",
            {"TIME": 4.393722000e9, "GEN": -30.0, "ENTG": 1.528048035348e7, "PWB": 0.0},
        ),
        (
            "GOFT_A1162_T2.csv",
            {"TIME": 3.06639400e9, "GEN": -27.5, "ENTG": 1.40141971e7, "PWB": 0.0},
        ),
    ],
)
def test_history(filename, data_ref):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", filename)
    data = toughio.read_history(filename)

    for k, v in data_ref.items():
        assert numpy.allclose(v, data[k].sum())


@pytest.mark.parametrize(
    "filename, filename_ref",
    [
        ("OUTPUT_ELEME.csv", "SAVE.out"),
        ("OUTPUT_ELEME.tec", "SAVE.out"),
        ("OUTPUT_ELEME_PETRASIM.csv", "SAVE.out"),
        ("OUTPUT.out", "SAVE.out"),
        ("OUTPUT_6.out", "SAVE_6.out"),
    ],
)
def test_output_eleme(filename, filename_ref):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", filename)
    outputs = toughio.read_output(filename)

    filename = os.path.join(this_dir, "support_files", "outputs", filename_ref)
    save = toughio.read_output(filename)

    assert len(outputs) == 5

    times_ref = [
        0.2592000e08,
        0.3155800e08,
        0.1577900e09,
        0.3155800e09,
        0.7889400e09,
    ]
    keys_ref = ["POR", "PRES", "SAT_G", "TEMP", "X", "Y", "Z"]
    for output, time_ref in zip(outputs, times_ref):
        assert time_ref == output.time
        assert (
            save.labels.tolist() == output.labels.tolist()
            if output.format in {"csv", "petrasim", "tough"}
            else output.labels == None
        )
        if output.format != "tough":
            assert keys_ref == sorted(list(output.data.keys()))

    assert numpy.allclose(save.data["X1"], outputs[-1].data["PRES"])
    assert numpy.allclose(save.data["X2"], outputs[-1].data["TEMP"], atol=0.1)


@pytest.mark.parametrize(
    "filename", ["OUTPUT_CONNE.csv", "OUTPUT.out", "OUTPUT_6.out"],
)
def test_output_conne(filename):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", filename)
    outputs = toughio.read_output(filename, connection=True)

    times_ref = [
        0.2592000e08,
        0.3155800e08,
        0.1577900e09,
        0.3155800e09,
        0.7889400e09,
    ]
    data_ref = [
        52542.0,
        52475.0,
        51146.0,
        49600.0,
        45623.0,
    ]
    for output, time_ref, data in zip(outputs, times_ref, data_ref):
        assert time_ref == output.time
        assert (
            len(set("".join(labels) for labels in output.labels))
            == output.data["HEAT"].size
        )
        assert numpy.allclose(data, numpy.abs(output.data["HEAT"]).mean(), atol=1.0)


@pytest.mark.parametrize(
    "output_ref, file_format",
    [
        (helpers.output_eleme, "csv"),
        (helpers.output_eleme[0], "csv"),
        (helpers.output_eleme, "petrasim"),
        (helpers.output_eleme[0], "petrasim"),
        (helpers.output_eleme, "tecplot"),
        (helpers.output_eleme[0], "tecplot"),
        (helpers.output_conne, "csv"),
        (helpers.output_conne[0], "csv"),
    ],
)
def test_output(output_ref, file_format):
    output = write_read(
        output=output_ref, writer_kws={"file_format": file_format}, reader_kws={},
    )

    output_ref = output_ref if isinstance(output_ref, list) else [output_ref]
    for out_ref, out in zip(output_ref, output):
        helpers.allclose_output(out_ref, out)


def test_save():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", "SAVE.out")
    save = toughio.read_output(filename)

    x_ref = [6.35804123e05, 1.42894499e02, 9.91868799e-01]
    assert numpy.allclose(
        x_ref, numpy.mean([save.data["X1"], save.data["X2"], save.data["X3"]], axis=1)
    )

    assert numpy.allclose(0.01, save.data["porosity"].mean())

    assert "userx" not in save.data.keys()



from __future__ import with_statement

from functools import partial

import numpy

from ...._common import get_label_length
from ...input.tough._helpers import read_record, str2float
from .._common import to_output

__all__ = [
    "read",
]


def read(filename, file_type, file_format, labels_order, label_length=None):
    """Read standard TOUGH OUTPUT."""
    with open(filename, "r") as f:
        headers, times, variables = _read_table(f, file_type, label_length)

        ilab = 1 if file_type == "element" else 2
        headers = headers[ilab + 1 :]
        labels = [[v[:ilab] for v in variable] for variable in variables]
        labels = (
            [[l[0] for l in label] for label in labels]
            if file_type == "element"
            else labels
        )
        variables = numpy.array(
            [[v[ilab + 1 :] for v in variable] for variable in variables]
        )

    return to_output(
        file_type, file_format, labels_order, headers, times, labels, variables
    )


def _read_table(f, file_type, label_length):
    """Read data table for current time step."""
    labels_key = "ELEM." if file_type == "element" else "ELEM1"

    first = True
    times, variables = [], []
    for line in f:
        line = line.strip()

        # Look for "TOTAL TIME"
        if line.startswith("TOTAL TIME"):
            # Read time step in following line
            line = next(f).strip()
            times.append(float(line.split()[0]))
            variables.append([])

            # Look for "ELEM." or "ELEM1"
            while True:
                try:
                    line = next(f).strip()
                    if line.startswith(labels_key):
                        break

                except StopIteration:
                    raise ValueError("No data related to {}s found.".format(file_type))

            # Read headers
            headers = line.split()

            # Look for next non-empty line
            while True:
                line = next(f)
                if line.strip():
                    break

            # Loop until end of output block
            while True:
                if line[:10].strip() and not line.strip().startswith("ELEM"):
                    line = line.lstrip()

                    if first and not label_length:
                        label_length = get_label_length(line[:9])
                        iend = (
                            label_length
                            if file_type == "element"
                            else 2 * label_length + 2
                        )

                    tmp = (
                        [line[:label_length]]
                        if file_type == "element"
                        else [line[:label_length], line[label_length + 2 : iend]]
                    )

                    line = line[iend:]
                    if first:
                        try:
                            # Set line parser and try parsing first line
                            reader = lambda line: [str2float(x) for x in line.split()]
                            _ = reader(line)

                        except ValueError:
                            # Determine number of characters for index
                            idx = line.replace("-", " ").split()[0]
                            nidx = line.index(idx) + len(idx)
                            ifmt = "{}s".format(nidx)

                            # Determine number of characters between two Es
                            i1 = line.find("E")
                            i2 = line.find("E", i1 + 1)

                            # Initialize data format
                            fmt = [ifmt]
                            if i2 >= 0:
                                di = i2 - i1
                                dfmt = "{}.{}e".format(di, di - 7)
                                fmt += 20 * [dfmt]  # Read 20 data columns at most
                            else:
                                fmt += ["12.5e"]
                            fmt = ",".join(fmt)

                            # Set line parser
                            reader = partial(read_record, fmt=fmt)

                        finally:
                            first = False

                    tmp += reader(line)
                    variables[-1].append([x for x in tmp if x is not None])

                line = next(f)
                if line[1:].startswith("@@@@@"):
                    break

    return headers, times, variables



import helpers
import numpy
import pytest

import toughio


@pytest.mark.parametrize(
    "nodal_distance, bound, coord",
    [
        ("line", True, False),
        ("line", False, False),
        ("orthogonal", True, False),
        ("orthogonal", False, False),
        ("line", True, True),
        ("line", False, True),
        ("orthogonal", True, True),
        ("orthogonal", False, True),
    ],
)
def test_mesh(nodal_distance, bound, coord):
    # Create 3D mesh
    dx = numpy.arange(3) + 1
    dy = numpy.arange(4) + 1
    dz = numpy.arange(5) + 1
    mesh = toughio.meshmaker.structured_grid(
        dx, dy, dz, material=helpers.random_string(5)
    )

    idx = numpy.random.choice(mesh.n_cells, mesh.n_cells // 2, replace=False)
    mesh.cell_data["material"][idx] = 2
    mesh.field_data[helpers.random_string(5)] = numpy.array([2, 3])

    idx = numpy.random.choice(mesh.n_cells, mesh.n_cells // 2, replace=False)
    boundary_condition = (
        (numpy.random.rand(mesh.n_cells) < numpy.random.rand()).astype(int)
        if bound
        else numpy.zeros(mesh.n_cells)
    )
    mesh.add_cell_data("boundary_condition", boundary_condition)

    parameters = helpers.write_read(
        filename="MESH",
        obj=None,
        writer=mesh.write_tough,
        reader=toughio.read_mesh,
        writer_kws={"nodal_distance": nodal_distance, "coord": coord},
        reader_kws={"file_format": "tough"},
    )

    # Check block ELEME
    assert sorted(mesh.labels) == sorted(parameters["elements"].keys())

    materials = [parameters["elements"][label]["material"] for label in mesh.labels]
    assert mesh.materials.tolist() == materials

    volumes = [
        parameters["elements"][label]["volume"]
        for label, bcond in zip(mesh.labels, boundary_condition)
        if not bcond
    ]
    assert numpy.allclose(mesh.volumes[boundary_condition == 0], volumes)

    volumes = [
        parameters["elements"][label]["volume"]
        for label, bcond in zip(mesh.labels, boundary_condition)
        if bcond
    ]
    assert numpy.allclose(mesh.volumes[boundary_condition == 1] * 1.0e50, volumes)

    centers = [parameters["elements"][label]["center"] for label in mesh.labels]
    assert numpy.allclose(mesh.centers, centers)

    # Check block COORD
    assert parameters["coordinates"] == coord

    # Check block CONNE
    nx, ny, nz = len(dx), len(dy), len(dz)
    lx, ly, lz = sum(dx), sum(dy), sum(dz)

    isot_x = [
        v["permeability_direction"] == 1 for v in parameters["connections"].values()
    ]
    assert sum(isot_x) == (nx - 1) * ny * nz

    isot_y = [
        v["permeability_direction"] == 2 for v in parameters["connections"].values()
    ]
    assert sum(isot_y) == nx * (ny - 1) * nz

    isot_z = [
        v["permeability_direction"] == 3 for v in parameters["connections"].values()
    ]
    assert sum(isot_z) == nx * ny * (nz - 1)

    interface_areas = [v["interface_area"] for v in parameters["connections"].values()]
    assert (
        sum(interface_areas)
        == (nx - 1) * ly * lz + (ny - 1) * lx * lz + (nz - 1) * lx * ly
    )

    angles = [
        v["gravity_cosine_angle"] == 0.0 for v in parameters["connections"].values()
    ]
    assert sum(angles) == sum(isot_x) + sum(isot_y)

    angles = [
        v["gravity_cosine_angle"] == 1.0 for v in parameters["connections"].values()
    ]
    assert sum(angles) == sum(isot_z)

    if not bound:
        xmin, ymin, zmin = mesh.centers.min(axis=0)
        xmax, ymax, zmax = mesh.centers.max(axis=0)
        distances_ref = (
            ny * nz * (xmax - xmin) + nx * nz * (ymax - ymin) + nx * ny * (zmax - zmin)
        )
        distances = [v["nodal_distances"] for v in parameters["connections"].values()]
        assert numpy.allclose(distances_ref, numpy.sum(distances))


@pytest.mark.parametrize("anisotropic", [True, False, None])
def test_incon(anisotropic):
    # Create 3D mesh
    dx = numpy.arange(3) + 1
    dy = numpy.arange(4) + 1
    dz = numpy.arange(5) + 1
    mesh = toughio.meshmaker.structured_grid(
        dx, dy, dz, material=helpers.random_string(5)
    )

    initial_condition = numpy.random.rand(mesh.n_cells, 4)
    mesh.add_cell_data("initial_condition", initial_condition)

    porosity = numpy.random.rand(mesh.n_cells)
    mesh.add_cell_data("porosity", porosity)

    if anisotropic is not None:
        permeability = (
            numpy.random.rand(mesh.n_cells, 3)
            if anisotropic
            else numpy.random.rand(mesh.n_cells)
        )
        mesh.add_cell_data("permeability", permeability)

    parameters = helpers.write_read(
        filename="INCON",
        obj=None,
        writer=mesh.write_incon,
        reader=toughio.read_mesh,
        reader_kws={"file_format": "tough"},
    )

    # Check block INCON
    assert sorted(mesh.labels) == sorted(parameters["initial_conditions"].keys())

    values = [
        parameters["initial_conditions"][label]["values"] for label in mesh.labels
    ]
    assert numpy.allclose(mesh.cell_data["initial_condition"], values)

    porosity = [
        parameters["initial_conditions"][label]["porosity"] for label in mesh.labels
    ]
    assert numpy.allclose(mesh.cell_data["porosity"], porosity)

    if anisotropic is not None:
        userx = numpy.array(
            [parameters["initial_conditions"][label]["userx"] for label in mesh.labels]
        )
        permeability = userx[:, :3] if anisotropic else userx[:, 0]
        assert numpy.allclose(mesh.cell_data["permeability"], permeability, atol=1.0e-4)



from __future__ import with_statement

import numpy

from .._common import to_output

__all__ = [
    "read",
    "write",
]


def read(filename, file_type, file_format, labels_order):
    """Read Petrasim OUTPUT_ELEME.csv."""
    with open(filename, "r") as f:
        # Headers
        line = f.readline().strip()
        headers = [header.strip() for header in line.split(",")[3:]]

        # Data
        times, elements, data = [], [], []
        while True:
            line = f.readline().strip()

            if line:
                line = line.split(",")
                times.append(float(line[0]))
                elements.append(line[1].strip())
                data.append([float(x) for x in line[3:]])
            else:
                break

    times = numpy.array(times)
    elements = numpy.array(elements)
    data = numpy.array(data)

    labels, unique_times, variables = [], [], []
    for time in numpy.unique(times):
        idx = times == time
        labels.append(elements[idx])
        unique_times.append(time)
        variables.append(data[idx])

    return to_output(
        file_type, file_format, labels_order, headers, unique_times, labels, variables
    )


def write(filename, output):
    """Write Petrasim OUTPUT_ELEME.csv."""
    out = output[-1]
    headers = []
    headers += ["X"] if "X" in out.data.keys() else []
    headers += ["Y"] if "Y" in out.data.keys() else []
    headers += ["Z"] if "Z" in out.data.keys() else []
    headers += [k for k in out.data.keys() if k not in {"X", "Y", "Z"}]

    with open(filename, "w") as f:
        # Headers
        record = ",".join(
            "{:>18}".format(header)
            for header in ["TIME [sec]", "ELEM", "INDEX"] + headers
        )
        f.write("{}\n".format(record))

        # Data
        for out in output:
            data = numpy.transpose([out.data[k] for k in headers])
            formats = ["{:20.12e}", "{:>18}", "{:20d}"]
            formats += ["{:20.12e}"] * len(out.data)

            i = 0
            for d in data:
                tmp = [out.time, out.labels[i], i + 1]
                tmp += [x for x in d]
                record = ",".join(fmt.format(x) for fmt, x in zip(formats, tmp))
                f.write("{}\n".format(record))
                i += 1



__all__ = [
    "save2incon",
]


def save2incon(argv=None):
    import os

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Check that SAVE file exists
    if not os.path.isfile(args.infile):
        raise ValueError("SAVE file '{}' not found.".format(args.infile))

    # Write INCON file
    with open(args.outfile, "w") as f:
        first = True
        with open(args.infile, "r") as fsave:
            for line in fsave:
                if first:
                    if not line.startswith("INCON"):
                        raise ValueError("Invalid SAVE file.")
                    else:
                        first = False
                if args.reset and line.startswith("+++"):
                    break
                f.write(line)

        if args.reset:
            f.write("\n")


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=("Convert a SAVE file to an INCON file."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Input file
    parser.add_argument(
        "infile", type=str, help="SAVE file to convert",
    )

    # Output file
    parser.add_argument(
        "outfile", type=str, help="converted INCON file",
    )

    # Reset
    parser.add_argument(
        "--reset", "-r", default=False, action="store_true", help="reset all counters",
    )

    return parser



from ._base import BaseRelativePermeability

__all__ = [
    "Corey",
]


class Corey(BaseRelativePermeability):

    _id = 3
    _name = "Corey"

    def __init__(self, slr, sgr):
        """
        Corey's curve.

        After Corey (1954).

        Parameters
        ----------
        slr : scalar
            Irreducible liquid saturation (RP(1)).
        sgr : scalar
            Irreducible gas saturation (RP(2)).

        """
        if slr + sgr >= 1.0:
            raise ValueError()
        self.parameters = [slr, sgr]

    def _eval(self, sl, *args):
        """Corey's curve."""
        slr, sgr = args
        sg = 1.0 - sl
        if sg < sgr:
            kl = 1.0
            kg = 0.0
        elif sg >= 1.0 - slr:
            kl = 0.0
            kg = 1.0
        else:
            Shat = (sl - slr) / (1.0 - slr - sgr)
            kl = Shat ** 4
            kg = (1.0 - Shat ** 2) * (1.0 - Shat) ** 2
        return kl, kg

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._slr, self._sgr]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 2:
            raise ValueError()
        self._slr = value[0]
        self._sgr = value[1]



from ._base import BaseRelativePermeability

__all__ = [
    "FattKlikoff",
]


class FattKlikoff(BaseRelativePermeability):

    _id = 7
    _name = "Fatt-Klikoff"

    def __init__(self, slr):
        """
        Fatt and Klikoff's function.

        After Fatt and Klikoff (1959).

        Parameters
        ----------
        slr : scalar
            Irreducible liquid saturation (RP(1)).

        """
        if slr >= 1.0:
            raise ValueError()
        self.parameters = [slr]

    def _eval(self, sl, *args):
        """Fatt and Klikoff's function."""
        (slr,) = args
        Seff = (sl - slr) / (1.0 - slr) if sl > slr else 0.0
        kl = Seff ** 3
        kg = (1.0 - Seff) ** 3
        return kl, kg

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._slr]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 1:
            raise ValueError()
        self._slr = value[0]



import sys
from copy import deepcopy

import helpers
import numpy
import pytest

import toughio


@pytest.mark.skipif(sys.version_info < (3,), reason="Order of keys in dictionary")
def test_time_series():
    # Import hybrid mesh
    mesh_ref = deepcopy(helpers.hybrid_mesh)

    # Create random time series
    num_steps = 5
    point_data_ref = [
        {"points": numpy.random.rand(mesh_ref.n_points)} for _ in range(num_steps)
    ]
    cell_data_ref = [
        {"cells": numpy.random.rand(mesh_ref.n_cells)} for i in range(num_steps)
    ]
    time_steps_ref = numpy.sort(numpy.random.rand(num_steps))

    # Write and read back XDMF
    filepath = helpers.tempdir("test.xdmf")
    toughio.write_time_series(
        filepath,
        mesh_ref.points,
        mesh_ref.cells,
        point_data_ref,
        cell_data_ref,
        time_steps_ref,
    )
    mesh = toughio.read_time_series(filepath)
    points, cells, point_data, cell_data, time_steps = mesh

    # Compare with reference data
    assert numpy.allclose(points, mesh_ref.points)

    for cell_ref, cell in zip(mesh_ref.cells, cells):
        assert cell_ref.type == cell.type
        assert numpy.allclose(cell_ref.data, cell.data)

    for t, pdata in enumerate(point_data):
        for k, v in pdata.items():
            assert numpy.allclose(v, point_data_ref[t][k])

    for t, cdata in enumerate(cell_data):
        for k, v in cdata.items():
            assert numpy.allclose(v, cell_data_ref[t][k])

    assert numpy.allclose(time_steps, time_steps_ref)



import numpy

from .._mesh._mesh import CellBlock, Mesh

__all__ = [
    "structured_grid",
]


def structured_grid(dx, dy, dz=None, origin=None, layer=False, material="dfalt"):
    """
    Generate 2D or 3D non-uniform structured grid.

    Parameters
    ----------
    dx : array_like
        Grid spacing along X axis.
    dy : array_like
        Grid spacing along Y axis.
    dz : array_like or None, optional, default None
        Grid spacing along Z axis. If `None`, generate 2D grid.
    origin : array_like or None, optional, default None
        Origin point coordinate.
    layer : bool, optional, default False
        If `True`, mesh will be generated by layers.
    material : str, optional, default 'dfalt'
        Default material name.

    Returns
    -------
    toughio.Mesh
        Output non-uniform structured mesh.

    """
    if not isinstance(dx, (list, tuple, numpy.ndarray)):
        raise TypeError()
    if not isinstance(dy, (list, tuple, numpy.ndarray)):
        raise TypeError()
    if not (dz is None or isinstance(dz, (list, tuple, numpy.ndarray))):
        raise TypeError()
    if not isinstance(material, str):
        raise TypeError()

    order = "F" if layer else "C"
    if dz is None:
        ndim = 2
        points, cells = _grid_2d(dx, dy, order)
    else:
        ndim = 3
        points, cells = _grid_3d(dx, dy, dz, order)

    if not (
        origin is None
        or (isinstance(origin, (list, tuple, numpy.ndarray)) and len(origin) == ndim)
    ):
        raise ValueError()
    origin = (
        numpy.asarray(origin)
        if origin is not None
        else (
            numpy.zeros(ndim) if ndim == 2 else numpy.array([0.0, 0.0, -numpy.sum(dz)])
        )
    )
    points += origin

    points = (
        points if ndim == 3 else numpy.column_stack((points, numpy.zeros(len(points))))
    )

    mesh = Mesh(points, cells)
    mesh.add_cell_data("material", numpy.ones(mesh.n_cells, dtype=int))
    mesh.field_data[material] = numpy.array([1, 3])

    return mesh


def _grid_3d(dx, dy, dz, order):
    """Generate 3D structured grid."""
    # Internal functions
    def meshgrid(x, y, z, indexing="ij", order=order):
        X, Y, Z = numpy.meshgrid(x, y, z, indexing=indexing)
        return X.ravel(order), Y.ravel(order), Z.ravel(order)

    def mesh_vertices(i, j, k):
        return [
            [i, j, k],
            [i + 1, j, k],
            [i + 1, j + 1, k],
            [i, j + 1, k],
            [i, j, k + 1],
            [i + 1, j, k + 1],
            [i + 1, j + 1, k + 1],
            [i, j + 1, k + 1],
        ]

    # Grid
    nx, ny, nz = len(dx), len(dy), len(dz)
    xyz_shape = [nx + 1, ny + 1, nz + 1]
    ijk_shape = [nx, ny, nz]
    X, Y, Z = meshgrid(*[numpy.cumsum(numpy.r_[0, ar]) for ar in [dx, dy, dz]])
    I, J, K = meshgrid(*[numpy.arange(n) for n in ijk_shape])

    # Points and cells
    points = [[x, y, z] for x, y, z in zip(X, Y, Z)]
    cells = [
        [
            numpy.ravel_multi_index(vertex, xyz_shape, order=order)
            for vertex in mesh_vertices(i, j, k)
        ]
        for i, j, k in zip(I, J, K)
    ]

    # Reorder cells from top to bottom
    if order == "F":
        cells = cells[::-1]

    n1 = nz if order == "F" else nx * ny
    n2 = nx * ny if order == "F" else nz
    for i in range(n1):
        i1 = i * n2
        i2 = i1 + n2
        cells[i1:i2] = cells[i2 - 1 : i1 - 1 : -1] if i > 0 else cells[i2 - 1 :: -1]

    return (
        numpy.array(points, dtype=float),
        [CellBlock("hexahedron", numpy.array(cells))],
    )


def _grid_2d(dx, dy, order):
    """Generate 2D structured grid."""
    # Internal functions
    def meshgrid(x, y, indexing="ij", order=order):
        X, Y = numpy.meshgrid(x, y, indexing=indexing)
        return X.ravel(order), Y.ravel(order)

    def mesh_vertices(i, j):
        return [
            [i, j],
            [i + 1, j],
            [i + 1, j + 1],
            [i, j + 1],
        ]

    # Grid
    nx, ny = len(dx), len(dy)
    xy_shape = [nx + 1, ny + 1]
    ij_shape = [nx, ny]
    X, Y = meshgrid(*[numpy.cumsum(numpy.r_[0, ar]) for ar in [dx, dy]])
    I, J = meshgrid(*[numpy.arange(n) for n in ij_shape])

    # Points and cells
    points = [[x, y] for x, y in zip(X, Y)]
    cells = [
        [
            numpy.ravel_multi_index(vertex, xy_shape, order=order)
            for vertex in mesh_vertices(i, j)
        ]
        for i, j in zip(I, J)
    ]

    return numpy.array(points, dtype=float), [CellBlock("quad", numpy.array(cells))]



from abc import ABCMeta, abstractmethod, abstractproperty

import numpy

__all__ = [
    "BaseRelativePermeability",
]


# See <https://stackoverflow.com/questions/35673474/using-abc-abcmeta-in-a-way-it-is-compatible-both-with-python-2-7-and-python-3-5>
ABC = ABCMeta("ABC", (object,), {"__slots__": ()})


class BaseRelativePermeability(ABC):

    _id = None
    _name = ""

    def __init__(self, *args):
        """
        Base class for relative permeability models.

        Do not use.

        """
        pass

    def __repr__(self):
        """Display relative permeability model informations."""
        out = [
            "{} relative permeability model (IRP = {}):".format(self._name, self._id)
        ]
        out += [
            "    RP({}) = {}".format(i + 1, parameter)
            for i, parameter in enumerate(self.parameters)
        ]
        return "\n".join(out)

    def __call__(self, sl):
        """Calculate relative permeability given liquid saturation."""
        if numpy.ndim(sl) == 0:
            if not (0.0 <= sl <= 1.0):
                raise ValueError()
            return self._eval(sl, *self.parameters)
        else:
            sl = numpy.asarray(sl)
            if not numpy.logical_and((sl >= 0.0).all(), (sl <= 1.0).all()):
                raise ValueError()
            return numpy.transpose([self._eval(sat, *self.parameters) for sat in sl])

    @abstractmethod
    def _eval(self, sl, *args):
        raise NotImplementedError()

    def plot(self, n=100, ax=None, figsize=(10, 8), plt_kws=None):
        """
        Plot relative permeability curve.

        Parameters
        ----------
        n : int, optional, default 100
            Number of saturation points.
        ax : matplotlib.pyplot.Axes or None, optional, default None
            Matplotlib axes. If `None`, a new figure and axe is created.
        figsize : array_like or None, optional, default None
            New figure size if `ax` is `None`.
        plt_kws : dict or None, optional, default None
            Additional keywords passed to :func:`matplotlib.pyplot.plot`.

        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError(
                "Plotting relative permeability curve requires matplotlib to be installed."
            )

        if not (isinstance(n, int) and n > 1):
            raise TypeError()
        if not (ax is None or isinstance(ax, plt.Axes)):
            raise TypeError()
        if not (figsize is None or isinstance(figsize, (tuple, list, numpy.ndarray))):
            raise TypeError()
        if len(figsize) != 2:
            raise ValueError()
        if not (plt_kws is None or isinstance(plt_kws, dict)):
            raise TypeError()

        # Plot parameters
        plt_kws = plt_kws if plt_kws is not None else {}
        _kwargs = {"linestyle": "-", "linewidth": 2}
        _kwargs.update(plt_kws)

        # Initialize figure
        if ax:
            ax1 = ax
        else:
            figsize = figsize if figsize else (8, 5)
            fig = plt.figure(figsize=figsize, facecolor="white")
            ax1 = fig.add_subplot(1, 1, 1)

        # Calculate liquid and gas relative permeability
        sl = numpy.linspace(0.0, 1.0, n)
        kl, kg = self(sl)

        # Plot
        ax1.plot(sl, kl, label="Liquid", **_kwargs)
        ax1.plot(sl, kg, label="Gas", **_kwargs)
        ax1.set_xlim(0.0, 1.0)
        ax1.set_ylim(0.0, 1.0)
        ax1.set_xlabel("Saturation (liquid)")
        ax1.set_ylabel("Relative permeability")
        ax1.grid(True, linestyle=":")

        plt.draw()
        plt.show()
        return ax1

    @property
    def id(self):
        """Return relative permeability model ID in TOUGH."""
        return self._id

    @property
    def name(self):
        """Return relative permeability model name."""
        return self._name

    @abstractproperty
    def parameters(self):
        raise NotImplementedError()

    @parameters.setter
    def parameters(self, value):
        raise NotImplementedError()



import os

import matplotlib.pyplot as plt
import numpy
import pytest

import toughio

if not os.environ.get("DISPLAY", ""):
    plt.switch_backend("Agg")


@pytest.mark.parametrize(
    "model, parameters, sl",
    [
        (toughio.capillarity.Linear, [1.0e6, 0.25, 0.4], None),
        (toughio.capillarity.Pickens, [1.0e6, 0.3, 1.3, 0.8], None),
        (
            toughio.capillarity.TRUST,
            [1.0e6, 0.3, 1.3, 0.8, 1.0e7],
            numpy.linspace(0.02, 1.0, 50),
        ),
        (toughio.capillarity.Milly, [0.25], None),
        (
            toughio.capillarity.vanGenuchten,
            [0.457, 0.0, 5.105e-4, 1.0e7, 1.0],
            numpy.linspace(0.02, 1.0, 50),
        ),
    ],
)
def test_capillarity(model, parameters, sl, monkeypatch):
    import json

    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "capillarity_references.json")
    with open(filename, "r") as f:
        pcap_ref = json.load(f)

    sl = numpy.linspace(0.0, 1.0, 51) if sl is None else sl

    cap = model(*parameters)
    pcap = cap(sl)
    assert numpy.allclose(pcap, pcap_ref[cap.name][: len(pcap)])

    monkeypatch.setattr(plt, "show", lambda: None)
    cap.plot()

    print(cap)



import os

import numpy

import toughio


def test_mesh():
    """Reference values are calculated in FLAC3D."""
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "all_cell_types.f3grid")
    mesh = toughio.read_mesh(filename)

    volumes = mesh.volumes.sum()
    assert numpy.allclose(volumes, 1.7083333333333344)

    face_areas = sum(face.sum() for face in mesh.face_areas)
    assert numpy.allclose(face_areas, 39.16535341331142)

    face_normals = sum(numpy.abs(face).sum() for face in mesh.face_normals)
    assert numpy.allclose(face_normals, 719.3660094744944)

    connections = mesh.connections.sum()
    assert numpy.allclose(connections, 25353)


def test_quality():
    dx = numpy.ones(3)
    dy = numpy.ones(3)
    dz = numpy.ones(3)
    mesh = toughio.meshmaker.structured_grid(dx, dy, dz, origin=numpy.zeros(3))

    assert numpy.allclose(mesh.qualities, numpy.ones(mesh.n_cells))



import logging
from itertools import chain

import numpy


def _materials(mesh):
    """Return materials of cell in mesh."""
    if "material" in mesh.cell_data.keys():
        if mesh.field_data:
            out = mesh.cell_data["material"]
            try:
                field_data_dict = {v[0]: k for k, v in mesh.field_data.items()}
                out = [field_data_dict[mat] for mat in out]
            except KeyError:
                logging.warning(
                    (
                        "field_data is not defined for all materials. "
                        "Returns materials as integers."
                    )
                )
                out = mesh.cell_data["material"]
        else:
            out = mesh.cell_data["material"]
    else:
        out = numpy.ones(mesh.n_cells, dtype=int)

    return numpy.asarray(out)


def _faces(mesh):
    """Return connectivity of faces of cell in mesh."""
    meshio_type_to_faces = {
        "tetra": {
            "triangle": numpy.array([[1, 2, 3], [0, 3, 2], [0, 1, 3], [0, 2, 1]]),
        },
        "pyramid": {
            "triangle": numpy.array([[0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4]]),
            "quad": numpy.array([[0, 3, 2, 1]]),
        },
        "wedge": {
            "triangle": numpy.array([[0, 2, 1], [3, 4, 5]]),
            "quad": numpy.array([[0, 1, 4, 3], [1, 2, 5, 4], [0, 3, 5, 2]]),
        },
        "hexahedron": {
            "quad": numpy.array(
                [
                    [0, 3, 2, 1],
                    [4, 5, 6, 7],
                    [0, 1, 5, 4],
                    [1, 2, 6, 5],
                    [2, 3, 7, 6],
                    [0, 4, 7, 3],
                ]
            ),
        },
    }

    return [
        [c[v] for v in meshio_type_to_faces[cell.type].values()]
        for cell in mesh.cells
        for c in cell.data
    ]


def _face_normals(mesh):
    """Return normal vectors of faces in mesh."""
    faces_dict, faces_cell, _ = _get_faces(_faces(mesh))

    # Face normal vectors
    normals = numpy.concatenate(
        [_get_triangle_normals(mesh, v) for k, v in faces_dict.items()]
    )
    normals_mag = numpy.linalg.norm(normals, axis=-1)
    normals /= normals_mag[:, None]

    # Reorganize outputs
    face_normals = [[] for _ in range(mesh.n_cells)]
    iface = numpy.concatenate([v for v in faces_cell.values()])
    for i, normal in zip(iface, normals):
        face_normals[i].append(normal)

    return face_normals


def _face_areas(mesh):
    """Return areas of faces in mesh."""
    faces_dict, faces_cell, _ = _get_faces(_faces(mesh))

    # Face areas
    areas = numpy.concatenate(
        [_get_triangle_normals(mesh, v) for k, v in faces_dict.items()]
    )
    areas = numpy.linalg.norm(areas, axis=-1)
    if "quad" in faces_dict.keys() and len(faces_dict["quad"]):
        tmp = numpy.concatenate(
            [
                _get_triangle_normals(mesh, v, [0, 2, 3])
                if k == "quad"
                else numpy.zeros((len(v), 3))
                for k, v in faces_dict.items()
            ]
        )
        areas += numpy.linalg.norm(tmp, axis=-1)
    areas *= 0.5

    # Reorganize outputs
    face_areas = [[] for _ in range(mesh.n_cells)]
    iface = numpy.concatenate([v for v in faces_cell.values()])
    for i, area in zip(iface, areas):
        face_areas[i].append(area)

    return [numpy.array(face) for face in face_areas]


def _volumes(mesh):
    """Return volumes of cell in mesh."""
    meshio_type_to_tetra = {
        "tetra": numpy.array([[0, 1, 2, 3]]),
        "pyramid": numpy.array([[0, 1, 3, 4], [1, 2, 3, 4]]),
        "wedge": numpy.array([[0, 1, 2, 5], [0, 1, 4, 5], [0, 3, 4, 5]]),
        "hexahedron": numpy.array(
            [[0, 1, 3, 4], [1, 4, 5, 6], [1, 2, 3, 6], [3, 4, 6, 7], [1, 3, 4, 6]]
        ),
    }

    out = []
    for cell in mesh.cells:
        tetras = numpy.vstack([c[meshio_type_to_tetra[cell.type]] for c in cell.data])
        tetras = mesh.points[tetras]
        out.append(
            numpy.sum(
                numpy.split(
                    numpy.abs(
                        _scalar_triple_product(
                            tetras[:, 1] - tetras[:, 0],
                            tetras[:, 2] - tetras[:, 0],
                            tetras[:, 3] - tetras[:, 0],
                        )
                    ),
                    len(cell.data),
                ),
                axis=1,
            )
            / 6.0
        )
    return numpy.concatenate(out)


def _connections(mesh):
    """
    Return mesh connections.

    Assume conformity and that points and cells are uniquely defined in mesh.

    """
    if numpy.shape(mesh.points)[1] != 3:
        raise ValueError("Connections for 2D mesh has not been implemented yet.")

    faces_dict, faces_cell, faces_index = _get_faces(_faces(mesh))
    faces_dict = {k: numpy.sort(numpy.vstack(v), axis=1) for k, v in faces_dict.items()}

    # Prune duplicate faces
    uf, tmp1, tmp2 = {}, {}, {}
    for k, v in faces_dict.items():
        up, uf[k] = numpy.unique(v, axis=0, return_inverse=True)
        tmp1[k] = [[] for _ in range(len(up))]
        tmp2[k] = [[] for _ in range(len(up))]

    # Make connections
    for k, v in uf.items():
        for i, j in enumerate(v):
            tmp1[k][j].append(faces_cell[k][i])
            tmp2[k][j].append(faces_index[k][i])
    conne = [vv for v in tmp1.values() for vv in v if len(vv) == 2]
    iface = [vv for v in tmp2.values() for vv in v if len(vv) == 2]

    # Reorganize output
    out = numpy.full((mesh.n_cells, 6), -1)
    for (i1, i2), (j1, j2) in zip(conne, iface):
        out[i1, j1] = i2
        out[i2, j2] = i1

    return out


def _qualities(mesh):
    """Return quality of cells as a measure of the orthogonality of its connections."""
    nodes = mesh.centers
    connections = mesh.connections
    face_normals = mesh.face_normals

    cell_list = set()
    centers, int_normals, mapper = [], [], []
    for i, connection in enumerate(connections):
        for iface, j in enumerate(connection):
            if j >= 0 and j not in cell_list:
                centers.append([nodes[i], nodes[j]])
                int_normals.append(face_normals[i][iface])
                mapper.append((i, j))

        cell_list.add(i)

    int_normals = numpy.array(int_normals)
    lines = numpy.diff(centers, axis=1)[:, 0]
    lines /= numpy.linalg.norm(lines, axis=1)[:, None]
    angles = numpy.abs((lines * int_normals).sum(axis=1))

    # Reorganize output
    out = [[] for _ in range(mesh.n_cells)]
    for (i, j), angle in zip(mapper, angles):
        out[i].append(angle)
        out[j].append(angle)

    return out


def _get_faces(faces):
    """Return dictionary of faces."""
    faces_dict = {"triangle": [], "quad": []}
    faces_cell = {"triangle": [], "quad": []}
    faces_index = {"triangle": [], "quad": []}
    numvert_to_face_type = {3: "triangle", 4: "quad"}

    for i, face in enumerate(faces):
        for j, f in enumerate(chain.from_iterable(face)):
            n = len(f)
            face_type = numvert_to_face_type[n]
            faces_dict[face_type].append(f[:n])
            faces_cell[face_type].append(i)
            faces_index[face_type].append(j)

    # Stack arrays or remove empty cells
    faces_dict = {k: numpy.vstack(v) for k, v in faces_dict.items() if len(v)}
    faces_cell = {k: v for k, v in faces_cell.items() if len(v)}
    faces_index = {k: v for k, v in faces_index.items() if len(v)}

    return faces_dict, faces_cell, faces_index


def _get_triangle_normals(mesh, faces, islice=None):
    """Calculate normal vectors of triangular faces."""
    islice = islice if islice is not None else [0, 1, 2]

    triangles = numpy.vstack([c[islice] for c in faces])
    triangles = mesh.points[triangles]

    return _cross(triangles[:, 1] - triangles[:, 0], triangles[:, 2] - triangles[:, 0])


def _cross(a, b):
    """Calculate cross product (faster than :func:`numpy.cross`)."""
    return a[:, [1, 2, 0]] * b[:, [2, 0, 1]] - a[:, [2, 0, 1]] * b[:, [1, 2, 0]]


def _scalar_triple_product(a, b, c):
    """Calculate determinant as scalar triple product."""
    return (a * _cross(b, c)).sum(axis=-1)



from ..._common import filetype_from_filename, register_format

__all__ = [
    "register",
    "read",
    "write",
]


_extension_to_filetype = {}
_reader_map = {}
_writer_map = {}


def register(file_format, extensions, reader, writer=None):
    """
    Register a new input format.

    Parameters
    ----------
    file_format : str
        File format to register.
    extensions : array_like
        List of extensions to associate to the new format.
    reader : callable
        Read fumction.
    writer : callable or None, optional, default None
        Write function.

    """
    register_format(
        fmt=file_format,
        ext_to_fmt=_extension_to_filetype,
        reader_map=_reader_map,
        writer_map=_writer_map,
        extensions=extensions,
        reader=reader,
        writer=writer,
    )


def read(filename, file_format=None, **kwargs):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str
        Input file name.
    file_format : str ('tough', 'json') or None, optional, default None
        Input file format.

    Other Parameters
    ----------------
    label_length : int or None, optional, default None
        Only if ``file_format = "tough"``. Number of characters in cell labels.

    Returns
    -------
    dict
        TOUGH input parameters.

    Note
    ----
    If ``file_format == 'tough'``, can also read `MESH`, `INCON` and `GENER` files.

    """
    if not isinstance(filename, str):
        raise TypeError()
    if not (file_format is None or file_format in {"tough", "json"}):
        raise ValueError()

    fmt = (
        file_format
        if file_format
        else filetype_from_filename(filename, _extension_to_filetype)
    )
    fmt = fmt if fmt else "tough"

    return _reader_map[fmt](filename, **kwargs)


def write(filename, parameters, file_format=None, **kwargs):
    """
    Write TOUGH input file.

    Parameters
    ----------
    filename : str
        Output file name.
    parameters : dict
        Parameters to export.
    file_format : str ('tough', 'json') or None, optional, default None
        Output file format.

    Other Parameters
    ----------------
    block : str {'all', 'gener', 'mesh', 'incon'}, optional, default 'all'
        Only if ``file_format = "tough"``. Blocks to be written:
         - 'all': write all blocks,
         - 'gener': only write block GENER,
         - 'mesh': only write blocks ELEME, COORD and CONNE,
         - 'incon': only write block INCON.

    """
    if not isinstance(filename, str):
        raise TypeError()
    if not isinstance(parameters, dict):
        raise TypeError()
    if not (file_format is None or file_format in {"tough", "json"}):
        raise ValueError()

    fmt = (
        file_format
        if file_format
        else filetype_from_filename(filename, _extension_to_filetype)
    )
    fmt = fmt if fmt else "tough"

    _writer_map[fmt](filename, parameters, **kwargs)



from ._base import BaseCapillarity

__all__ = [
    "TRUST",
]


class TRUST(BaseCapillarity):

    _id = 3
    _name = "TRUST"

    def __init__(self, p0, slr, eta, pe, pmax):
        """
        TRUST capillary pressure.

        After Narasimhan et al. (1978).

        Parameters
        ----------
        p0 : scalar
            Capillary pressure strength (CP(1)).
        slr : scalar
            Irreducible liquid saturation (CP(2)).
        eta : scalar
            CP(3).
        pe : scalar
            Capillary entry pressure (CP(4)).
        pmax : scalar
            Maximum pressure (CP(5)).

        """
        if slr < 0.0:
            raise ValueError()
        if eta == 0.0:
            raise ValueError()
        self.parameters = [p0, slr, eta, pe, pmax]

    def _eval(self, sl, *args):
        """TRUST capillary pressure."""
        p0, slr, eta, pe, pmax = args
        if sl > slr:
            pcap = -pe - p0 * ((1.0 - sl) / (sl - slr)) ** (1.0 / eta)
        else:
            pcap = -abs(pmax)
        pcap = max(pcap, -abs(pmax))
        pcap *= (1.0 - sl) / 0.001 if sl > 0.999 else 1.0
        return pcap

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._p0, self._slr, self._eta, self._pe, self._pmax]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 5:
            raise ValueError()
        self._p0 = value[0]
        self._slr = value[1]
        self._eta = value[2]
        self._pe = value[3]
        self._pmax = value[4]



import os
import sys

from setuptools import find_packages, setup

base_dir = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(base_dir, "toughio", "__about__.py"), "rb") as f:
    exec (f.read(), about)


DISTNAME = "toughio"
DESCRIPTION = "Pre- and post-processing Python library for TOUGH"
LONG_DESCRIPTION = open("README.rst").read()
VERSION = about["__version__"]
AUTHOR = about["__author__"]
AUTHOR_EMAIL = about["__author_email__"]
URL = about["__website__"]
LICENSE = about["__license__"]
CLASSIFIERS = [
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: BSD License",
    "Natural Language :: English",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: MacOS",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
]

# TODO: use new environment markers when dropping support for python < 3
# REQUIREMENTS = [
#     "meshio == 2.3.10; python_version <= '2.7'",
#     "meshio >= 3.0.0; python_version > '2.7'",
# ]
# EXTRA_REQUIREMENTS = {
#     "full": [
#         "pyvista == 0.22.4; python_version <= '2.7'",
#         "pyvista >= 0.23.1; python_version > '2.7'",
#     ],
# }
if sys.version_info < (3,):
    REQUIREMENTS = ["meshio == 2.3.10"]
    EXTRA_REQUIREMENTS = {"full": ["imageio==2.6.1", "lxml", "pyvista == 0.22.4"]}
else:
    REQUIREMENTS = ["meshio >= 4.0.11, < 5.0"]
    EXTRA_REQUIREMENTS = {"full": ["pyvista >= 0.23.1"]}
REQUIREMENTS += ["numpy >= 1.13.0"]
EXTRA_REQUIREMENTS["full"] += ["matplotlib", "h5py", "netcdf4", "scipy>=0.9"]


PACKAGE_DATA = {
    "toughio.data": ["CO2TAB"],
}

ENTRY_POINTS = {
    "console_scripts": [
        "toughio-co2tab = toughio._cli:co2tab",
        "toughio-export = toughio._cli:export",
        "toughio-extract = toughio._cli:extract",
        "toughio-merge = toughio._cli:merge",
        "toughio-save2incon = toughio._cli:save2incon",
    ],
}


if __name__ == "__main__":
    setup(
        name=DISTNAME,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/x-rst",
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        license=LICENSE,
        install_requires=REQUIREMENTS,
        extras_require=EXTRA_REQUIREMENTS,
        python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*",
        classifiers=CLASSIFIERS,
        version=about["__version__"],
        packages=find_packages() + ["toughio.data"],
        package_data=PACKAGE_DATA,
        entry_points=ENTRY_POINTS,
    )



from __future__ import division, with_statement

from ...._common import block_to_format, get_label_length
from ._helpers import prune_nones_dict, prune_nones_list, read_record

__all__ = [
    "read",
]


def read(filename, label_length=None):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str
        Input file name.
    label_length : int or None, optional, default None
        Number of characters in cell labels.

    Returns
    -------
    dict
        TOUGH input parameters.

    """
    if not (label_length is None or isinstance(label_length, int)):
        raise TypeError()
    if isinstance(label_length, int) and not 5 <= label_length < 10:
        raise ValueError()

    with open(filename, "r") as f:
        out = read_buffer(f, label_length)

    return out


def read_buffer(f, label_length):
    """Read TOUGH input file."""
    parameters = {}

    # Title
    line = f.readline().strip()
    if line[:5] not in {"ROCKS", "ELEME", "INCON", "GENER"}:
        title = [line]
        while True:
            line = f.readline().strip()
            if not line.startswith("ROCKS"):
                title.append(line)
            else:
                break

        parameters["title"] = title[0] if len(title) == 1 else title
    f.seek(0)

    # Loop over blocks
    # Some blocks (INCON, INDOM, PARAM) need to rewind to previous line but tell and seek are disabled by next
    # See <https://stackoverflow.com/questions/22688505/is-there-a-way-to-go-back-when-reading-a-file-using-seek-and-calls-to-next>
    fiter = iter(f.readline, "")
    for line in fiter:
        if line.startswith("ROCKS"):
            parameters.update(_read_rocks(fiter))
        elif line.startswith("RPCAP"):
            rpcap = _read_rpcap(fiter)
            if "default" in parameters.keys():
                parameters["default"].update(rpcap)
            else:
                parameters["default"] = rpcap
        elif line.startswith("FLAC"):
            flac = _read_flac(fiter, parameters["rocks_order"])
            parameters["flac"] = flac["flac"]
            for k, v in flac["rocks"].items():
                parameters["rocks"][k].update(v)
        elif line.startswith("MULTI"):
            parameters.update(_read_multi(fiter))
        elif line.startswith("SOLVR"):
            parameters.update(_read_solvr(fiter))
        elif line.startswith("START"):
            parameters["start"] = True
        elif line.startswith("PARAM"):
            param = _read_param(fiter, f)
            parameters["options"] = param["options"]
            parameters["extra_options"] = param["extra_options"]
            if "default" in parameters.keys():
                parameters["default"].update(param["default"])
            else:
                parameters["default"] = param["default"]
        elif line.startswith("SELEC"):
            parameters.update(_read_selec(fiter))
        elif line.startswith("INDOM"):
            indom = _read_indom(fiter, f)
            for k, v in indom["rocks"].items():
                parameters["rocks"][k].update(v)
        elif line.startswith("MOMOP"):
            parameters.update(_read_momop(fiter))
        elif line.startswith("TIMES"):
            parameters.update(_read_times(fiter))
        elif line.startswith("FOFT"):
            parameters.update(_read_oft(fiter, "element_history"))
        elif line.startswith("COFT"):
            parameters.update(_read_oft(fiter, "connection_history"))
        elif line.startswith("GOFT"):
            parameters.update(_read_oft(fiter, "generator_history"))
        elif line.startswith("GENER"):
            parameters.update(_read_gener(fiter, label_length))
        elif line.startswith("DIFFU"):
            parameters.update(_read_diffu(fiter, f))
        elif line.startswith("OUTPU"):
            parameters.update(_read_outpu(fiter))
        elif line.startswith("ELEME"):
            parameters.update(_read_eleme(fiter, label_length))
            parameters["coordinates"] = False
        elif line.startswith("COORD"):
            coord = _read_coord(fiter)
            for k, v in zip(parameters["elements_order"], coord):
                parameters["elements"][k]["center"] = v
            parameters["coordinates"] = True
        elif line.startswith("CONNE"):
            parameters.update(_read_conne(fiter, label_length))
        elif line.startswith("INCON"):
            parameters.update(_read_incon(fiter, label_length, f))
        elif line.startswith("NOVER"):
            parameters["nover"] = True
        elif line.startswith("ENDCY"):
            break

    return parameters


def _read_rocks(f):
    """Read ROCKS block data."""
    fmt = block_to_format["ROCKS"]
    rocks = {"rocks": {}, "rocks_order": []}

    while True:
        line = next(f)

        if line.strip():
            # Record 1
            data = read_record(line, fmt[1])
            rock = data[0]
            rocks["rocks"][rock] = {
                "density": data[2],
                "porosity": data[3],
                "permeability": data[4] if len(set(data[4:7])) == 1 else data[4:7],
                "conductivity": data[7],
                "specific_heat": data[8],
            }

            nad = data[1]
            if nad is not None:
                # Record 2
                line = next(f)
                data = read_record(line, fmt[2])
                rocks["rocks"][rock].update(
                    {
                        "compressibility": data[0],
                        "expansivity": data[1],
                        "conductivity_dry": data[2],
                        "tortuosity": data[3],
                        "klinkenberg_parameter": data[4],
                        "distribution_coefficient_3": data[5],
                        "distribution_coefficient_4": data[6],
                    }
                )

            if nad and nad > 1:
                rocks["rocks"][rock].update(_read_rpcap(f))

            rocks["rocks_order"].append(rock)
        else:
            break

    rocks["rocks"] = {k: prune_nones_dict(v) for k, v in rocks["rocks"].items()}

    return rocks


def _read_rpcap(f):
    """Read RPCAP block data."""
    fmt = block_to_format["RPCAP"]
    rpcap = {}

    for key in ["relative_permeability", "capillarity"]:
        line = next(f)
        data = read_record(line, fmt)
        if data[0] is not None:
            rpcap[key] = {
                "id": data[0],
                "parameters": prune_nones_list(data[2:]),
            }

    return rpcap


def _read_flac(f, rocks_order):
    """Read FLAC block data."""
    fmt = block_to_format["FLAC"]
    flac = {"rocks": {}, "flac": {}}

    # Record 1
    line = next(f)
    data = read_record(line, fmt[1])
    flac["flac"]["creep"] = data[0]
    flac["flac"]["porosity_model"] = data[1]
    flac["flac"]["version"] = data[2]

    # Additional records
    for rock in rocks_order:
        flac["rocks"][rock] = {}

        line = next(f)
        data = read_record(line, fmt[2])
        flac["rocks"][rock]["permeability_model"] = {
            "id": data[0],
            "parameters": prune_nones_list(data[1:]),
        }

        line = next(f)
        data = read_record(line, fmt[3])
        flac["rocks"][rock]["equivalent_pore_pressure"] = {
            "id": data[0],
            "parameters": prune_nones_list(data[2:]),
        }

    flac["flac"] = prune_nones_dict(flac["flac"])

    return flac


def _read_multi(f):
    """Read MULTI block data."""
    fmt = block_to_format["MULTI"]
    multi = {}

    line = next(f)
    data = read_record(line, fmt)
    multi["n_component"] = data[0]
    multi["isothermal"] = data[1] == data[0]
    multi["n_phase"] = data[2]

    return multi


def _read_solvr(f):
    """Read SOLVR block data."""
    fmt = block_to_format["SOLVR"]
    solvr = {}

    line = next(f)
    data = read_record(line, fmt)
    solvr["solver"] = {
        "method": data[0],
        "z_precond": data[2],
        "o_precond": data[4],
        "rel_iter_max": data[5],
        "eps": data[6],
    }

    return solvr


def _read_param(f, fh):
    """Read PARAM block data."""
    fmt = block_to_format["PARAM"]
    param = {}

    # Record 1
    line = next(f)
    data = read_record(line, fmt[1])
    param["options"] = {
        "n_iteration": data[0],
        "verbosity": data[1],
        "n_cycle": data[2],
        "n_second": data[3],
        "n_cycle_print": data[4],
        "temperature_dependence_gas": data[7],
        "effective_strength_vapor": data[8],
    }
    param["extra_options"] = {
        i + 1: int(x) for i, x in enumerate(data[5]) if x.isdigit()
    }

    # Record 2
    line = next(f)
    data = read_record(line, fmt[2])
    param["options"].update(
        {
            "t_ini": data[0],
            "t_max": data[1],
            "t_steps": data[2],
            "t_step_max": data[3],
            "gravity": data[5],
            "t_reduce_factor": data[6],
            "mesh_scale_factor": data[7],
        }
    )

    t_steps = int(data[2])
    if t_steps >= 0.0:
        param["options"]["t_steps"] = t_steps
    else:
        param["options"]["t_steps"] = []
        for _ in range(-t_steps):
            line = next(f)
            data = read_record(line, fmt[3])
            param["options"]["t_steps"] += prune_nones_list(data)
        if len(param["options"]["t_steps"]) == 1:
            param["options"]["t_steps"] = param["options"]["t_steps"][0]

    # Record 3
    line = next(f)
    data = read_record(line, fmt[4])
    param["options"].update(
        {
            "eps1": data[0],
            "eps2": data[1],
            "w_upstream": data[3],
            "w_newton": data[4],
            "derivative_factor": data[5],
        }
    )

    # Record 4 and record 5 (EOS7R)
    line = next(f)
    data = read_record(line, fmt[5])

    i = fh.tell()
    try:
        line = next(f)
        data += read_record(line, fmt[5])
    except ValueError:
        fh.seek(i)

    if any(x is not None for x in data):
        data = prune_nones_list(data)
        param["default"] = {"initial_condition": data}
    else:
        param["default"] = {}

    # Remove Nones
    param["options"] = prune_nones_dict(param["options"])
    param["extra_options"] = prune_nones_dict(param["extra_options"])

    return param


def _read_selec(f):
    """Read SELEC block data."""
    fmt = block_to_format["SELEC"]
    selec = {"selections": {}}

    line = next(f)
    data = read_record(line, fmt[1])
    selec["selections"]["integers"] = {k + 1: v for k, v in enumerate(data)}

    if selec["selections"]["integers"][1]:
        selec["selections"]["floats"] = []
        for _ in range(selec["selections"]["integers"][1]):
            line = next(f)
            data = read_record(line, fmt[2])
            selec["selections"]["floats"].append(prune_nones_list(data))

    selec["selections"]["integers"] = prune_nones_dict(selec["selections"]["integers"])
    if selec["selections"]["integers"][1] == 1:
        selec["selections"]["floats"] = selec["selections"]["floats"][0]

    return selec


def _read_indom(f, fh):
    """Read INDOM block data."""
    fmt = block_to_format["INDOM"]
    indom = {"rocks": {}}

    line = next(f)
    two_lines = True
    while True:
        if line.strip():
            # Record 1
            rock = read_record(line, fmt[5])[0]

            # Record 2
            line = next(f)
            data = read_record(line, fmt[0])

            # Record 3 (EOS7R)
            if two_lines:
                i = fh.tell()
                line = next(f)

                if line.strip():
                    try:
                        data += read_record(line, fmt[0])
                    except ValueError:
                        two_lines = False
                        fh.seek(i)
                else:
                    fh.seek(i)

            data = prune_nones_list(data)
            indom["rocks"][rock] = {"initial_condition": data}
        else:
            break

        line = next(f)

    return indom


def _read_momop(f):
    """Read MOMOP block data."""
    fmt = block_to_format["MOMOP"]

    line = next(f)
    data = read_record(line, fmt)
    momop = {
        "more_options": {i + 1: int(x) for i, x in enumerate(data[0]) if x.isdigit()}
    }

    return momop


def _read_times(f):
    """Read TIMES block data."""
    fmt = block_to_format["TIMES"]
    times = {"times": []}

    # Record 1
    line = next(f)
    data = read_record(line, fmt[1])
    n_times = data[0]

    # Record 2
    while len(times["times"]) < n_times:
        line = next(f)
        data = read_record(line, fmt[2])
        times["times"] += prune_nones_list(data)

    if n_times == 1:
        times["times"] = times["times"][0]

    return times


def _read_oft(f, oft):
    """Read FOFT, COFT and GOFT blocks data."""
    history = {oft: []}

    while True:
        line = next(f).rstrip()

        if line:
            history[oft].append(line)
        else:
            break

    return history


def _read_gener(f, label_length):
    """Read GENER block data."""
    fmt = block_to_format["GENER"]
    gener = {"generators": {}}

    line = next(f)
    if not label_length:
        label_length = get_label_length(line[:9])

    while True:
        if line.strip():
            data = read_record(line, fmt[label_length])
            label = data[0]
            tmp = {
                "name": [data[1]],
                "nseq": [data[2]],
                "nadd": [data[3]],
                "nads": [data[4]],
                "type": [data[7]],
                "layer_thickness": [data[11]],
            }

            ltab = data[5]
            if ltab and ltab > 1:
                itab = data[8]
                keys = ["times", "rates"]
                keys += ["specific_enthalpy"] if itab else []
                for key in keys:
                    table = []

                    while len(table) < ltab:
                        line = next(f)
                        data = read_record(line, fmt[0])
                        table += prune_nones_list(data)

                    tmp[key] = [table]
            else:
                tmp.update(
                    {
                        "times": [None],
                        "rates": [data[9]],
                        "specific_enthalpy": [data[10]],
                    }
                )

            if label in gener["generators"].keys():
                for k, v in gener["generators"][label].items():
                    v += tmp[k]
            else:
                gener["generators"][label] = tmp
        else:
            break

        line = next(f)

    # Tidy up
    for generator in gener["generators"].values():
        for k, v in generator.items():
            if len(v) == 1:
                generator[k] = v[0]
            else:
                if all(x is None for x in v):
                    generator[k] = None

    return {
        k: {kk: prune_nones_dict(vv) for kk, vv in v.items()} for k, v in gener.items()
    }


def _read_diffu(f, fh):
    """Read DIFFU block data."""
    fmt = block_to_format["DIFFU"]
    diffu = {"diffusion": []}

    while True:
        i = fh.tell()
        line = next(f)

        if line.split():
            try:
                data = read_record(line, fmt)
                diffu["diffusion"].append(prune_nones_list(data))
            except ValueError:
                fh.seek(i)
                break
        else:
            break

    return diffu


def _read_outpu(f):
    """Read OUTPU block data."""
    fmt = block_to_format["OUTPU"]
    outpu = {"output": {}}

    # Format
    line = next(f).strip()
    if line and not line.isdigit():
        outpu["output"]["format"] = line if line else None
        line = next(f).strip()

    # Variables
    names = {}
    if line.isdigit():
        num_vars = int(line)
        outpu["output"]["variables"] = {}

        for _ in range(num_vars):
            line = next(f)
            data = read_record(line, fmt[3])
            name = data[0].lower()

            tmp = prune_nones_list(data[1:])
            if name not in names.keys():
                names[name] = 1
                outpu["output"]["variables"][name] = tmp
            else:
                if names[name] == 1:
                    outpu["output"]["variables"][name] = [
                        outpu["output"]["variables"][name],
                        tmp,
                    ]
                else:
                    outpu["output"]["variables"][name].append(tmp)

                names[name] += 1

        for k, v in outpu["output"]["variables"].items():
            outpu["output"]["variables"][k] = (
                None if len(v) == 0 else v[0] if len(v) == 1 else v
            )

    return outpu


def _read_eleme(f, label_length):
    """Read ELEME block data."""
    fmt = block_to_format["ELEME"]
    eleme = {"elements": {}, "elements_order": []}

    line = next(f)
    if not label_length:
        label_length = get_label_length(line[:9])

    while True:
        if line.strip():
            data = read_record(line, fmt[label_length])
            label = data[0]
            rock = data[3].strip()
            eleme["elements"][label] = {
                "nseq": data[1],
                "nadd": data[2],
                "material": int(rock) if all(r.isdigit() for r in rock) else rock,
                "volume": data[4],
                "heat_exchange_area": data[5],
                "permeability_modifier": data[6],
                "center": data[7:10],
            }

            eleme["elements_order"].append(label)
        else:
            break

        line = next(f)

    eleme["elements"] = {k: prune_nones_dict(v) for k, v in eleme["elements"].items()}

    return eleme


def _read_coord(f):
    """Read COORD block data."""
    fmt = block_to_format["COORD"]
    coord = []

    line = next(f)
    while True:
        if line.strip():
            data = read_record(line, fmt)
            coord.append(data)
        else:
            break

        line = next(f)

    return coord


def _read_conne(f, label_length):
    """Read CONNE block data."""
    fmt = block_to_format["CONNE"]
    conne = {"connections": {}, "connections_order": []}

    line = next(f)
    if not label_length:
        label_length = get_label_length(line[:9])

    while True:
        if line.strip() and not line.startswith("+++"):
            data = read_record(line, fmt[label_length])
            label = data[0]
            conne["connections"][label] = {
                "nseq": data[1],
                "nadd": data[2:4],
                "permeability_direction": data[4],
                "nodal_distances": data[5:7],
                "interface_area": data[7],
                "gravity_cosine_angle": data[8],
                "radiant_emittance_factor": data[9],
            }

            conne["connections_order"].append(label)
        else:
            break

        line = next(f)

    conne["connections"] = {
        k: prune_nones_dict(v) for k, v in conne["connections"].items()
    }

    return conne


def _read_incon(f, label_length, fh):
    """Read INCON block data."""
    fmt = block_to_format["INCON"]
    incon = {"initial_conditions": {}, "initial_conditions_order": []}

    line = next(f)
    if not label_length:
        label_length = get_label_length(line[:9])

    two_lines = True
    while True:
        if line.strip() and not line.startswith("+++"):
            # Record 1
            data = read_record(line, fmt[label_length])
            label = data[0]
            userx = prune_nones_list(data[4:9])
            incon["initial_conditions"][label] = {
                "porosity": data[3],
                "userx": userx if userx else None,
            }

            # Record 2
            line = next(f)
            data = read_record(line, fmt[0])

            # Record 3 (EOS7R)
            if two_lines:
                i = fh.tell()
                line = next(f)

                if line.strip() and not line.startswith("+++"):
                    try:
                        data += read_record(line, fmt[0])
                    except ValueError:
                        two_lines = False
                        fh.seek(i)
                else:
                    fh.seek(i)

            incon["initial_conditions"][label]["values"] = prune_nones_list(data)
            incon["initial_conditions_order"].append(label)
        else:
            break

        line = next(f)

    incon["initial_conditions"] = {
        k: prune_nones_dict(v) for k, v in incon["initial_conditions"].items()
    }

    return incon



import collections

import numpy

__all__ = [
    "Output",
]


Output = collections.namedtuple("Output", ["type", "format", "time", "labels", "data"])


def to_output(file_type, file_format, labels_order, headers, times, labels, variables):
    """Create an Output namedtuple."""
    outputs = [
        Output(
            file_type,
            file_format,
            time,
            numpy.array(label),
            {k: v for k, v in zip(headers, numpy.transpose(variable))},
        )
        for time, label, variable in zip(times, labels, variables)
    ]
    return (
        [reorder_labels(out, labels_order) for out in outputs]
        if labels_order is not None and file_type == "element"
        else outputs
    )


def reorder_labels(data, labels):
    """Reorder output or save cell data according to input labels."""
    if len(data.labels) != len(labels):
        raise ValueError()

    mapper = {k: v for v, k in enumerate(data.labels)}
    idx = [mapper[label] for label in labels]
    data.labels[:] = data.labels[idx]

    for k, v in data.data.items():
        data.data[k] = v[idx]

    return data
